@@ -2,7 +2,7 @@
 #this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 import datetime
-from itertools import groupby
+from itertools import groupby, chain
 from functools import partial
 from trytond.model import Workflow, ModelView, ModelSQL, fields
 from trytond.modules.company import CompanyReport
@@ -12,10 +12,9 @@ from trytond.backend import TableHandler
 from trytond.pyson import If, Eval, Bool, PYSONEncoder, Id
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
-from trytond.config import CONFIG
 
 __all__ = ['Sale', 'SaleInvoice', 'SaleIgnoredInvoice', 'SaleRecreatedInvoice',
-    'SaleLine', 'SaleLineTax', 'SaleLineInvoiceLine', 'SaleLineIgnoredMove',
+    'SaleLine', 'SaleLineTax', 'SaleLineIgnoredMove',
     'SaleLineRecreatedMove', 'SaleReport', 'Template', 'Product',
     'ShipmentOut', 'ShipmentOutReturn', 'Move', 'OpenCustomer',
     'HandleShipmentExceptionAsk', 'HandleShipmentException',
@@ -167,19 +166,17 @@ class Sale(Workflow, ModelSQL, ModelView):
         super(Sale, cls).__setup__()
         cls._order.insert(0, ('sale_date', 'DESC'))
         cls._order.insert(1, ('id', 'DESC'))
-        cls._constraints += [
-            ('check_method', 'wrong_method')
-            ]
         cls._error_messages.update({
-                'wrong_method': 'Wrong combination of method!',
-                'addresses_required': 'Invoice and Shipment addresses must be '
-                'defined for the quotation.',
-                'warehouse_required': 'Warehouse must be defined for the ' \
-                    'quotation.',
-                'missing_account_receivable': 'It misses '
-                'an "Account Receivable" on the party "%s"!',
-                'delete_cancel': 'Sale "%s" must be cancelled ' \
-                    'before deletion!',
+                'invalid_method': ('Invalid combination of shipment and '
+                    'invoicing methods on sale "%s".'),
+                'addresses_required': ('Invoice and Shipment addresses must be '
+                    'defined for the quotation of sale "%s".'),
+                'warehouse_required': ('Warehouse must be defined for the '
+                    'quotation of sale "%s".'),
+                'missing_account_receivable': ('It misses '
+                        'an "Account Receivable" on the party "%s".'),
+                'delete_cancel': ('Sale "%s" must be cancelled before '
+                    'deletion.'),
                 })
         cls._transitions |= set((
                 ('draft', 'quotation'),
@@ -230,7 +227,7 @@ class Sale(Workflow, ModelSQL, ModelView):
     def __register__(cls, module_name):
         pool = Pool()
         SaleLine = pool.get('sale.line')
-        SaleLine_InvoiceLine = pool.get('sale.line-account.invoice.line')
+        sale_line_invoice_line_table = 'sale_line_invoice_lines_rel'
         Move = pool.get('stock.move')
         cursor = Transaction().cursor
         # Migration from 1.2: packing renamed into shipment
@@ -263,7 +260,7 @@ class Sale(Workflow, ModelSQL, ModelView):
         # state confirmed splitted into confirmed and processing
         if (TableHandler.table_exist(cursor, SaleLine._table)
                 and TableHandler.table_exist(cursor,
-                    SaleLine_InvoiceLine._table)
+                    sale_line_invoice_line_table)
                 and TableHandler.table_exist(cursor, Move._table)):
             # Wrap subquery inside an other inner subquery because MySQL syntax
             # doesn't allow update a table and select from the same table in a
@@ -277,12 +274,13 @@ class Sale(Workflow, ModelSQL, ModelView):
                         'FROM "%s" AS s '
                         'INNER JOIN "%s" AS l ON l.sale = s.id '
                         'LEFT JOIN "%s" AS li ON li.sale_line = l.id '
-                        'LEFT JOIN "%s" AS m ON m.sale_line = l.id '
+                        'LEFT JOIN "%s" AS m ON m.origin = \'%s,\' || l.id '
                         "WHERE s.state = 'confirmed' "
                             'AND (li.id IS NOT NULL '
                                 'OR m.id IS NOT NULL)) AS foo)'
                 % (cls._table, cls._table, SaleLine._table,
-                    SaleLine_InvoiceLine._table, Move._table))
+                    sale_line_invoice_line_table, Move._table,
+                    SaleLine.__name__))
 
         # Add index on create_date
         table = TableHandler(cursor, cls, module_name)
@@ -385,9 +383,10 @@ class Sale(Workflow, ModelSQL, ModelView):
         return res
 
     def on_change_with_party_lang(self, name=None):
+        Config = Pool().get('ir.configuration')
         if self.party and self.party.lang:
             return self.party.lang.code
-        return CONFIG['language']
+        return Config.get_language()
 
     def on_change_lines(self):
         pool = Pool()
@@ -495,21 +494,20 @@ class Sale(Workflow, ModelSQL, ModelView):
                     'invoice_state': state,
                     })
 
-    def get_shipments_returns(attribute):
+    def get_shipments_returns(model_name):
         "Computes the returns or shipments"
         def method(self, name):
-            shipments = []
+            Model = Pool().get(model_name)
+            shipments = set()
             for line in self.lines:
                 for move in line.moves:
-                    ship_or_return = getattr(move, attribute)
-                    if bool(ship_or_return):
-                        if ship_or_return.id not in shipments:
-                            shipments.append(ship_or_return.id)
-            return shipments
+                    if isinstance(move.shipment, Model):
+                        shipments.add(move.shipment.id)
+            return list(shipments)
         return method
 
-    get_shipments = get_shipments_returns('shipment_out')
-    get_shipment_returns = get_shipments_returns('shipment_out_return')
+    get_shipments = get_shipments_returns('stock.shipment.out')
+    get_shipment_returns = get_shipments_returns('stock.shipment.out.return')
 
     def get_moves(self, name):
         return [m.id for l in self.lines for m in l.moves]
@@ -537,17 +535,22 @@ class Sale(Workflow, ModelSQL, ModelView):
                     'shipment_state': state,
                     })
 
+    @classmethod
+    def validate(cls, sales):
+        super(Sale, cls).validate(sales)
+        for sale in sales:
+            sale.check_method()
+
     def check_method(self):
         '''
         Check the methods.
         '''
         if (self.invoice_method == 'shipment'
                 and self.shipment_method in ('invoice', 'manual')):
-            return False
+            self.raise_user_error('invalid_method', (self.rec_name,))
         if (self.shipment_method == 'invoice'
                 and self.invoice_method in ('shipment', 'manual')):
-            return False
-        return True
+            self.raise_user_error('invalid_method', (self.rec_name,))
 
     def get_rec_name(self, name):
         return (self.reference or str(self.id)
@@ -577,7 +580,7 @@ class Sale(Workflow, ModelSQL, ModelView):
 
     def check_for_quotation(self):
         if not self.invoice_address or not self.shipment_address:
-            self.raise_user_error('addresses_required')
+            self.raise_user_error('addresses_required', (self.rec_name,))
         for line in self.lines:
             if line.quantity >= 0:
                 location = line.from_location
@@ -586,7 +589,8 @@ class Sale(Workflow, ModelSQL, ModelView):
             if ((not location or not line.warehouse)
                     and line.product
                     and line.product.type in ('goods', 'assets')):
-                self.raise_user_error('warehouse_required')
+                self.raise_user_error('warehouse_required',
+                    (self.rec_name,))
 
     @classmethod
     def set_reference(cls, sales):
@@ -670,31 +674,21 @@ class Sale(Workflow, ModelSQL, ModelView):
         '''
         pool = Pool()
         Invoice = pool.get('account.invoice')
-        SaleLine = pool.get('sale.line')
         if self.invoice_method == 'manual':
             return
 
         if not self.party.account_receivable:
             self.raise_user_error('missing_account_receivable',
-                    error_args=(self.party.rec_name,))
+                (self.party.rec_name,))
 
         invoice_lines = self._get_invoice_line_sale_line(invoice_type)
         if not invoice_lines:
             return
 
         invoice = self._get_invoice_sale(invoice_type)
+        invoice.lines = list(chain.from_iterable(invoice_lines.itervalues()))
         invoice.save()
 
-        for line in self.lines:
-            if line.id not in invoice_lines:
-                continue
-            for invoice_line in invoice_lines[line.id]:
-                invoice_line.invoice = invoice.id
-                invoice_line.save()
-                SaleLine.write([line], {
-                        'invoice_lines': [('add', [invoice_line.id])],
-                        })
-
         with Transaction().set_user(0, set_context=True):
             Invoice.update_taxes([invoice])
 
@@ -737,7 +731,6 @@ class Sale(Workflow, ModelSQL, ModelView):
         Create and return shipments of type shipment_type
         '''
         pool = Pool()
-        SaleLine = pool.get('sale.line')
 
         if self.shipment_method == 'manual':
             return
@@ -747,11 +740,9 @@ class Sale(Workflow, ModelSQL, ModelView):
             return
         if shipment_type == 'out':
             keyfunc = partial(self._group_shipment_key, moves.values())
-            move_shipment_key = 'shipment_out'
             Shipment = pool.get('stock.shipment.out')
         elif shipment_type == 'return':
             keyfunc = partial(self._group_return_key, moves.values())
-            move_shipment_key = 'shipment_out_return'
             Shipment = pool.get('stock.shipment.out.return')
         moves = moves.items()
         moves = sorted(moves, key=keyfunc)
@@ -766,14 +757,10 @@ class Sale(Workflow, ModelSQL, ModelView):
                 }
             values.update(dict(key))
             with Transaction().set_user(0, set_context=True):
-                shipment = Shipment.create(values)
+                shipment = Shipment(**values)
+                shipment.moves = [x[1] for x in grouped_moves]
+                shipment.save()
                 shipments.append(shipment)
-                for line_id, move in grouped_moves:
-                    setattr(move, move_shipment_key, shipment.id)
-                    move.save()
-                    SaleLine.write([SaleLine(line_id)], {
-                            'moves': [('add', [move.id])],
-                            })
         if shipment_type == 'out':
             with Transaction().set_user(0, set_context=True):
                 Shipment.wait(shipments)
@@ -791,7 +778,7 @@ class Sale(Workflow, ModelSQL, ModelView):
         cls.cancel(sales)
         for sale in sales:
             if sale.state != 'cancel':
-                cls.raise_user_error('delete_cancel', sale.rec_name)
+                cls.raise_user_error('delete_cancel', (sale.rec_name,))
         super(Sale, cls).delete(sales)
 
     @classmethod
@@ -960,13 +947,16 @@ class SaleLine(ModelSQL, ModelView):
     description = fields.Text('Description', size=None, required=True)
     note = fields.Text('Note')
     taxes = fields.Many2Many('sale.line-account.tax', 'line', 'tax', 'Taxes',
-        domain=[('parent', '=', None)], states={
+        domain=[('parent', '=', None), ['OR',
+                ('group', '=', None),
+                ('group.kind', 'in', ['sale', 'both'])],
+            ],
+        states={
             'invisible': Eval('type') != 'line',
             }, depends=['type'])
-    invoice_lines = fields.Many2Many('sale.line-account.invoice.line',
-            'sale_line', 'invoice_line', 'Invoice Lines', readonly=True)
-    moves = fields.One2Many('stock.move', 'sale_line', 'Moves',
-            readonly=True)
+    invoice_lines = fields.One2Many('account.invoice.line', 'origin',
+        'Invoice Lines', readonly=True)
+    moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
     moves_ignored = fields.Many2Many('sale.line-ignored-stock.move',
             'sale_line', 'move', 'Ignored Moves', readonly=True)
     moves_recreated = fields.Many2Many('sale.line-recreated-stock.move',
@@ -981,11 +971,13 @@ class SaleLine(ModelSQL, ModelView):
     to_location = fields.Function(fields.Many2One('stock.location',
             'To Location'), 'get_to_location')
     delivery_date = fields.Function(fields.Date('Delivery Date',
-            on_change_with=['product', '_parent_sale.sale_date'],
+            on_change_with=['product', 'quantity', '_parent_sale.sale_date'],
             states={
-                'invisible': Eval('type') != 'line',
+                'invisible': ((Eval('type') != 'line')
+                    | (If(Bool(Eval('quantity')), Eval('quantity', 0), 0)
+                        <= 0)),
                 },
-            depends=['type']),
+            depends=['type', 'quantity']),
         'on_change_with_delivery_date')
 
     @classmethod
@@ -993,12 +985,13 @@ class SaleLine(ModelSQL, ModelView):
         super(SaleLine, cls).__setup__()
         cls._order.insert(0, ('sequence', 'ASC'))
         cls._error_messages.update({
-            'customer_location_required': 'The customer location is required!',
-            'missing_account_revenue': 'It misses '
-                    'an "Account Revenue" on product "%s"!',
-            'missing_account_revenue_property': 'It misses '
-                    'an "account Revenue" default property!',
-            })
+                'customer_location_required': ('Sale "%(sale)s" is missing the '
+                    'customer location in line "%(line)s".'),
+                'missing_account_revenue': ('Product "%(product)s" of sale '
+                    '%(sale)s misses a revenue account.'),
+                'missing_account_revenue_property': ('Sale "%(sale)s" '
+                    'misses an "account revenue" default property.'),
+                })
 
     @classmethod
     def __register__(cls, module_name):
@@ -1196,7 +1189,7 @@ class SaleLine(ModelSQL, ModelView):
                 return self.warehouse.input_location.id
 
     def on_change_with_delivery_date(self, name=None):
-        if self.product:
+        if self.product and self.quantity > 0:
             date = self.sale.sale_date if self.sale else None
             return self.product.compute_delivery_date(date=date)
 
@@ -1259,15 +1252,19 @@ class SaleLine(ModelSQL, ModelView):
         if self.product:
             invoice_line.account = self.product.account_revenue_used
             if not invoice_line.account:
-                self.raise_user_error('missing_account_revenue',
-                        error_args=(self.product.rec_name,))
+                self.raise_user_error('missing_account_revenue', {
+                        'sale': self.sale.rec_name,
+                        'product': self.product.rec_name,
+                        })
         else:
             for model in ('product.template', 'product.category'):
                 invoice_line.account = Property.get('account_revenue', model)
                 if invoice_line.account:
                     break
             if not invoice_line.account:
-                self.raise_user_error('missing_account_revenue_property')
+                self.raise_user_error('missing_account_revenue_property',
+                    (self.sale.rec_name,))
+        invoice_line.origin = self
         return [invoice_line]
 
     @classmethod
@@ -1303,7 +1300,7 @@ class SaleLine(ModelSQL, ModelView):
         else:
             quantity = 0.0
             for invoice_line in self.invoice_lines:
-                if invoice_line.invoice.state in ('open', 'paid'):
+                if invoice_line.invoice.state in ('posted', 'paid'):
                     quantity += Uom.compute_qty(invoice_line.unit,
                         invoice_line.quantity, self.unit)
 
@@ -1315,7 +1312,10 @@ class SaleLine(ModelSQL, ModelView):
         if quantity <= 0.0:
             return
         if not self.sale.party.customer_location:
-            self.raise_user_error('customer_location_required')
+            self.raise_user_error('customer_location_required', {
+                    'sale': self.sale.rec_name,
+                    'line': self.rec_name,
+                    })
         with Transaction().set_user(0, set_context=True):
             move = Move()
         move.quantity = quantity
@@ -1328,6 +1328,7 @@ class SaleLine(ModelSQL, ModelView):
         move.unit_price = self.unit_price
         move.currency = self.sale.currency
         move.planned_date = self.delivery_date
+        move.origin = self
         return move
 
 
@@ -1341,16 +1342,6 @@ class SaleLineTax(ModelSQL):
             select=True, required=True)
 
 
-class SaleLineInvoiceLine(ModelSQL):
-    'Sale Line - Invoice Line'
-    __name__ = 'sale.line-account.invoice.line'
-    _table = 'sale_line_invoice_lines_rel'
-    sale_line = fields.Many2One('sale.line', 'Sale Line', ondelete='CASCADE',
-            select=True, required=True)
-    invoice_line = fields.Many2One('account.invoice.line', 'Invoice Line',
-            ondelete='RESTRICT', select=True, required=True)
-
-
 class SaleLineIgnoredMove(ModelSQL):
     'Sale Line - Ignored Move'
     __name__ = 'sale.line-ignored-stock.move'
@@ -1511,25 +1502,23 @@ class ShipmentOut:
             for shipment in shipments:
                 move_ids.extend([x.id for x in shipment.outgoing_moves])
 
-            sale_lines = SaleLine.search([
-                    ('moves', 'in', move_ids),
-                    ])
-            if sale_lines:
-                for sale_line in sale_lines:
-                    if sale_line.sale not in sales:
-                        sales.append(sale_line.sale)
-
             with Transaction().set_user(0, set_context=True):
-                sales = Sale.browse([s.id for s in sales])
-            Sale.process(sales)
+                sale_lines = SaleLine.search([
+                        ('moves', 'in', move_ids),
+                        ])
+                if sale_lines:
+                    sales = list(set(l.sale for l in sale_lines))
+                    Sale.process(sales)
 
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
     def draft(cls, shipments):
+        SaleLine = Pool().get('sale.line')
         for shipment in shipments:
             for move in shipment.outgoing_moves:
-                if move.state == 'cancel' and move.sale_line:
+                if (move.state == 'cancel'
+                        and isinstance(move.origin, SaleLine)):
                     cls.raise_user_error('reset_move')
 
         return super(ShipmentOut, cls).draft(shipments)
@@ -1560,25 +1549,27 @@ class ShipmentOutReturn:
             for shipment in shipments:
                 move_ids.extend([x.id for x in shipment.incoming_moves])
 
-            sale_lines = SaleLine.search([
-                    ('moves', 'in', move_ids),
-                    ])
-            if sale_lines:
-                for sale_line in sale_lines:
-                    if sale_line.sale not in sales:
-                        sales.append(sale_line.sale)
-
             with Transaction().set_user(0, set_context=True):
-                sales = Sale.browse([s.id for s in sales])
-            Sale.process(sales)
+                sale_lines = SaleLine.search([
+                        ('moves', 'in', move_ids),
+                        ])
+                if sale_lines:
+                    for sale_line in sale_lines:
+                        if sale_line.sale not in sales:
+                            sales.append(sale_line.sale)
+
+                    sales = Sale.browse([s.id for s in sales])
+                    Sale.process(sales)
 
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
     def draft(cls, shipments):
+        SaleLine = Pool().get('sale.line')
         for shipment in shipments:
             for move in shipment.incoming_moves:
-                if move.state == 'cancel' and move.sale_line:
+                if (move.state == 'cancel'
+                        and isinstance(move.origin, SaleLine)):
                     cls.raise_user_error('reset_move')
 
         return super(ShipmentOutReturn, cls).draft(shipments)
@@ -1586,33 +1577,51 @@ class ShipmentOutReturn:
 
 class Move:
     __name__ = 'stock.move'
-    sale_line = fields.Many2One('sale.line', 'Sale Line', select=True,
-        states={
-            'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
     sale = fields.Function(fields.Many2One('sale.sale', 'Sale', select=True),
-            'get_sale', searcher='search_sale')
+        'get_sale', searcher='search_sale')
     sale_exception_state = fields.Function(fields.Selection([
         ('', ''),
         ('ignored', 'Ignored'),
         ('recreated', 'Recreated'),
         ], 'Exception State'), 'get_sale_exception_state')
 
+    @classmethod
+    def __register__(cls, module_name):
+        cursor = Transaction().cursor
+
+        super(Move, cls).__register__(module_name)
+
+        table = TableHandler(cursor, cls, module_name)
+
+        # Migration from 2.6: remove sale_line
+        if table.column_exist('sale_line'):
+            cursor.execute('UPDATE "' + cls._table + '" '
+                'SET origin = \'sale.line,\' || sale_line '
+                'WHERE sale_line IS NOT NULL')
+            table.drop_column('sale_line')
+
+    @classmethod
+    def _get_origin(cls):
+        models = super(Move, cls)._get_origin()
+        models.append('sale.line')
+        return models
+
     def get_sale(self, name):
-        if self.sale_line:
-            return self.sale_line.sale.id
+        SaleLine = Pool().get('sale.line')
+        if isinstance(self.origin, SaleLine):
+            return self.origin.sale.id
 
     @classmethod
     def search_sale(cls, name, clause):
-        return [('sale_line.' + name,) + clause[1:]]
+        return [('origin.' + name,) + tuple(clause[1:]) + ('sale.line',)]
 
     def get_sale_exception_state(self, name):
-        if not self.sale_line:
+        SaleLine = Pool().get('sale.line')
+        if not isinstance(self.origin, SaleLine):
             return ''
-        if self in self.sale_line.moves_recreated:
+        if self in self.origin.moves_recreated:
             return 'recreated'
-        if self in self.sale_line.moves_ignored:
+        if self in self.origin.moves_ignored:
             return 'ignored'
         return ''
 
@@ -1624,17 +1633,13 @@ class Move:
 
         super(Move, cls).write(moves, vals)
         if 'state' in vals and vals['state'] in ('cancel',):
-            sales = set()
-            sale_lines = SaleLine.search([
-                    ('moves', 'in', [m.id for m in moves]),
-                    ])
-            if sale_lines:
-                for sale_line in sale_lines:
-                    sales.add(sale_line.sale)
-            if sales:
-                with Transaction().set_user(0, set_context=True):
-                    sales = Sale.browse([s.id for s in sales])
-                Sale.process(sales)
+            with Transaction().set_user(0, set_context=True):
+                sale_lines = SaleLine.search([
+                        ('moves', 'in', [m.id for m in moves]),
+                        ])
+                if sale_lines:
+                    sales = list(set(l.sale for l in sale_lines))
+                    Sale.process(sales)
 
     @classmethod
     def delete(cls, moves):
@@ -1642,20 +1647,17 @@ class Move:
         Sale = pool.get('sale.sale')
         SaleLine = pool.get('sale.line')
 
-        sales = set()
-        sale_lines = SaleLine.search([
-                ('moves', 'in', [m.id for m in moves]),
-                ])
+        with Transaction().set_user(0, set_context=True):
+            sale_lines = SaleLine.search([
+                    ('moves', 'in', [m.id for m in moves]),
+                    ])
 
         super(Move, cls).delete(moves)
 
         if sale_lines:
-            for sale_line in sale_lines:
-                sales.add(sale_line.sale)
-            if sales:
-                with Transaction().set_user(0, set_context=True):
-                    sales = Sale.browse([s.id for s in sales])
-                Sale.process(list(sales))
+            sales = list(set(l.sale for l in sale_lines))
+            with Transaction().set_user(0, set_context=True):
+                Sale.process(sales)
 
 
 class OpenCustomer(Wizard):
@@ -1677,7 +1679,6 @@ class OpenCustomer(Wizard):
         model_data, = ModelData.search([
                 ('fs_id', '=', 'act_open_customer'),
                 ('module', '=', 'sale'),
-                ('inherit', '=', None),
                 ], limit=1)
         wizard = Wizard(model_data.db_id)
 
